import gamestate, thing
import pygame, math, random

class Drop:
  c = (64, 64, 128)
  def __init__(self, xrange, yrange, smin, smax):
    self.x = random.uniform(0, xrange)
    self.y = random.uniform(0, yrange)
    self.s = random.uniform(smin, smax)
  def addtop(self, xrange, y):
    self.x = random.uniform(0, xrange)
    self.y = y - self.s
  def draw(self, surf, dx, dy):
    pygame.draw.aaline(surf, self.c, (self.x, self.y), (self.x + dx, self.y + dy))
    self.x += self.s * dx / dy
    self.y += self.s

class Rain:
  def __init__(self, ndrops):
    self.ndrops = ndrops
    self.drops = [Drop(gamestate.sx, gamestate.sy, 30, 60) for j in range(ndrops)]
    self.dx, self.dy = -3, 40
    self.lframe = 0
  def draw(self, screen):
    for drop in self.drops:
      drop.draw(screen, self.dx, self.dy)
      if drop.y > gamestate.sy: drop.addtop(gamestate.sx, -self.dy)
  def strike(self, pos):
    self.lpos = pos
    self.lframe = 10
    x, y = pos
    r = 0.3
    for t in thing.Thing.tgroup:
      if t.movable:
        dx, dy = x - t.x, y - t.y
        if dx**2 + dy**2 < r**2:
          d = math.sqrt(dx**2 + dy**2)
          t.dropeverything()
          a = math.exp(-5*d) / 10.
          nx, ny = t.x, t.y
          for s in [8,7,6,5,4,3,2,1]:
            nx, ny = nx - dx * a*s, ny - dy * a*s
            t.doq.append("self.move0((%f,%f))" % (nx, ny))
      if isinstance(t, thing.Structure) and not t.inprogress:
        dx, dy = x - t.x, y - t.y
        if dx**2 + dy**2 < r**2:
          t.addhealth(-400)
  def skycolor(self):
    if self.lframe: self.lframe -= 1
    return ((0,64,0), (255,255,255))[self.lframe % 2]
  def drawlightning(self):
    if self.lframe:
      for j in range(3):
        x, y = gamestate.Pxy(self.lpos)
        while y >= 0:
          nx, ny = x + random.randint(-5, 20), y - 40
          pygame.draw.aaline(gamestate.screen, (255, 255, 255), (x, y), (nx, ny))
          x, y = nx, ny
      x, y = gamestate.Pxy(self.lpos)
      d = 8 * (10 - self.lframe)
      r = pygame.Rect(x - 2*d, y - d, 4*d, 2*d)
      pygame.draw.ellipse(gamestate.screen, (255, 255, 255), r, 2)



