import gamestate, data, imagebank
import pygame, math, random

# class Thing: all sprites in this game. Main functionality of this class
#   is to convert internal (x,y,z) coordinates into screen coordinates.
#   Also updates to the screen are done through Thing.group, so all sprites
#   should be included in this.
# class Structure: a building, has different distance requirements from
#   units and other buildings
# class Critter: a movable sprite, includes both units and enemies

# class Shadow: not a Thing. Why is it in this module?

class Thing (pygame.sprite.Sprite):
  group = pygame.sprite.LayeredUpdates()
  tgroup = pygame.sprite.Group()
  sgroup = pygame.sprite.Group()
  igroup = pygame.sprite.Group()
  dpx = dpy = None
  def __init__(self, p):
    pygame.sprite.Sprite.__init__(self)
    self.movable = False
    self.shadow = None
    self.indicator = None
    self.gots = []
    self.move0(p)
    self.trailz = []
    self.leader = None
    self.target = None
    if self.dpx == None:
      self.dpx, self.dpy = self.rect.width/2, self.rect.height
    self.doq = []   # The "do" queue
    Thing.group.add(self)
    Thing.tgroup.add(self)
  def addshadow(self):
    self.shadow = Shadow(self)
  def select(self):
    if self.indicator: return True
    self.indicator = Indicator(self)
    Thing.igroup.add(self)
  def unselect(self):
    if self.indicator:
      self.indicator.kill()
      self.indicator = None
    Thing.igroup.remove(self)
  def update(self):
    if len(self.doq):
      exec self.doq.pop(0)
      if not self.alive(): return
    layer = gamestate.Rxy(self.x,self.y)[1]
    Thing.group.change_layer(self, layer)
    x, y = gamestate.Pxy(self.x, self.y, self.z)
    self.rect.topleft = x - self.dpx, y - self.dpy
    if self.indicator:
      self.indicator.rect.midbottom = self.rect.midtop
      Thing.group.change_layer(self.indicator, layer)
    for g in self.gots: g.trail(self.r + g.r + 0.01)
  def drawpoint(self, color = (255, 128, 0)):
    x, y = gamestate.Pxy(self.x, self.y, self.z)
    gamestate.screen.set_at((int(x), int(y)), color)
    r = self.r * gamestate.kx
    rect = pygame.Rect(int(x-r), int(y-r/2.), int(2*r), int(r))
    pygame.draw.ellipse(gamestate.screen, color, rect, 1)
  def moving(self):
    return False
  def move0(self, p):  # instantaneous move
    if len(p) == 3:
      self.x, self.y, self.z = p
    else:
      self.x, self.y = p
      self.z = 0
    for g in self.gots: g.trail()
  move = move0
  def scootch(self, other):
    if self == other: return
    if self.moving() or other.moving(): return
    if not self.touching(other): return
    dx, dy = other.x - self.x, other.y - self.y
    r = other.r + self.r
    d = math.sqrt(dx * dx + dy * dy) / r
    if d < 0.01:
      p = random.uniform(0, 6.28)
      C, S = math.cos(p) * 1.1, math.sin(p) * 1.1
      if other.movable:
        self.move((self.x + self.r * C, self.y + self.r * S))
        other.move((other.x - other.r * C, other.y - other.r * S))
      else:
        self.move((self.x + r * C, self.y + r * S))
    else:
      a = (1.1 / d - 1)
      if other.movable:
        other.move((other.x + a * dx, other.y + a * dy))
      elif self.movable:
        self.move((self.x - a * dx, self.y - a * dy))
  def trail(self, distance = None):
    if not self.leader: return
    dx, dy, r = self.leader.x - self.x, self.leader.y - self.y, self.leader.r + self.r
    if distance == None: distance = r
    self.z = 0
    while len(self.trailz) > 1: self.z = self.trailz.pop(0)
    if dx**2 + dy**2 > r**2:
      d = math.sqrt(dx**2 + dy**2)
      self.move0((self.leader.x - dx*r/d, self.leader.y - dy*r/d, self.z))
    self.trailz.append(self.leader.z)
  def nab(self, item):
    self.gots.append(item)
    item.leader = self
  def use(self, target):
    if not self.gots: return
    self.target = target
    self.approachtarget()
  def approachtarget(self):
    if not self.target or not self.gots: return
    dx, dy = self.target.x - self.x, self.target.y - self.y
    d = math.sqrt(dx**2 + dy**2)
    r = 0.1
    if d > r:
      r *= 0.9
      x, y = self.target.x - dx * r / d, self.target.y - dy * r / d
      self.move((x,y))
      self.doq.append("self.approachtarget()")
    else:
      item = self.gots[0]
      dx, dy = self.target.x - item.x, self.target.y - item.y
      zs = parabola(0, 0.15, 0., 10) + [0,]
      for j in range(len(zs)):
        x, y, z = item.x + dx * j / len(zs), item.y + dy * j / len(zs), zs[j]
        item.doq.append("self.move0((%f, %f, %f))" % (x, y, z))
      item.doq.append("self.apply()")
      item.target = self.target
      self.loseitem()
      self.target = False
  def loseitem(self):
    if not self.gots: return
    item = self.gots.pop(0)
    for g in item.gots: g.leader = self
    self.gots += item.gots
    item.gots = []
    item.leader = None    
  def apply(self):
    pass
  def touching(self, other):
    dx, dy, r = other.x - self.x, other.y - self.y, other.r + self.r
    return dx * dx + dy * dy < r * r
  def dropeverything(self):
    del self.doq[:]
    if self.z:
      self.move0((self.x, self.y, 0))
  def kill(self):
    if self.shadow: self.shadow.kill()
    pygame.sprite.Sprite.kill(self)


def unselectall():
  for i in Thing.igroup.sprites(): i.unselect()

class Critter (Thing):
  group = pygame.sprite.Group()
  def __init__(self, p, img):
    self.image = img
    self.rect = self.image.get_rect()
    self.r = self.rect.width / gamestate.kx / 2.2
    Thing.__init__(self, p)
    self.movable = True
    self.addshadow()
    Critter.group.add(self)
  def moving(self):
    return len(self.doq) > 0
  def goavailable(self):
    Thing.tgroup.add(self)
    Thing.sgroup.add(self)
  def gounavailable(self):
    self.unselect()
    Thing.tgroup.remove(self)
    Thing.sgroup.remove(self)
  def update(self):
    if len(self.doq):
      Thing.update(self)
      if not self.alive(): return
      if not self.moving():
        for c in Thing.tgroup.sprites():
          self.scootch(c)
    else:
      Thing.update(self)
  def hopto(self, p1):
    x0, y0 = self.x, self.y
    x1, y1 = p1
    dx, dy = x1 - x0, y1 - y0
    d = math.sqrt(dx ** 2 + dy ** 2)
    n = int(d / self.speed) + 1
    njump = int(self.maxjump)
    c = []
    r0, r1 = 0, self.maxjump
    for a in [float(x)/n for x in range(1, n)]:
      x = x0 + dx * a
      y = y0 + dy * a
      if d*a > r1: r0, r1 = r1, r1 + self.maxjump
      if r1 > d: r1 = d
      z = (d*a - r0) * (r1 - d*a) * 3
      self.doq.append("self.move0((%f,%f,%f))" % (x,y,z))
    self.doq.append("self.move0((%f,%f,%f))" % (x1,y1,0))
  move = hopto
    
class Structure (Thing):
  group = pygame.sprite.Group()
  buildr = 0.08
  glowframes = 10
  bdhealth = 1
  fhealth = 10
  def __init__(self, p):
    self.rect = self.image.get_rect()
    self.r = self.rect.width / gamestate.kx / 2.2
    Thing.__init__(self, p)
    Structure.group.add(self)
    self.movable = False
    self.inprogress = True
    self.glowing = 0
    self.ibar = False
    self.health = 0
    if self.inprogress:
      Thing.tgroup.remove(self)
  def isdisabled(self):
    return self.health < self.maxhealth / 2. or self.inprogress

  def addhealth(self, dhealth):
    self.glowing += self.glowframes
    self.dhealth = dhealth / self.glowframes

  def update(self):

    if self.inprogress:
      self.health += self.bdhealth
      if self.health >= self.maxhealth:
        self.health = self.maxhealth
        self.inprogress = False
        Thing.tgroup.add(self)
        for t in Thing.tgroup.sprites(): self.scootch(t)
    if self.health < 0:
      if self.ibar: self.ibar.kill()
      self.kill()
      return

    if self.glowing > 0:
      iname, f = ("rglow", "glow")[self.dhealth > 0], min(self.glowing * 10, 100)
      self.image = self.imagebank.image("%s%d" % (iname, f))
      self.health += self.dhealth
      self.glowing -= 1
    elif self.inprogress:
      self.image = self.imagebank.image("ghost")
    elif self.isdisabled():
      self.image = self.imagebank.image("stony")
    else:
      self.image = self.imagebank.image()

    if self.health < self.maxhealth:
      a, b = int(self.health / self.fhealth), int(self.maxhealth / self.fhealth)
      if not self.ibar: self.ibar = Ibar(b)
      self.ibar.image = self.ibar.getimage(a, b)
      self.ibar.rect.midbottom = self.rect.midtop
    elif self.ibar:
      self.ibar.kill()
      self.ibar = False
      
    Thing.update(self)
    if self.ibar:
      layer = Thing.group.get_layer_of_sprite(self)
      Thing.group.change_layer(self.ibar, layer)
    

class Seeder (Structure):
  imagebank = imagebank.Imagebank("seeder")
  dpx, dpy = 10, 106
  maxhealth = 1200
  def __init__(self, p):
    self.image = Seeder.imagebank.image()
    Structure.__init__(self, p)
    self.health = self.maxhealth
    self.ticker = 0
  def update(self):
    if not self.isdisabled():
      self.ticker += 1.
      if self.ticker > 100 and len(Acorn.group) < gamestate.maxacorns:
        self.ticker = 0
        self.sprout()
    Structure.update(self)
  def sprout(self):
    s = Acorn((random.uniform(-.3, .3), random.uniform(-.3, .3)))

class Dewdropper (Structure):
  imagebank = imagebank.Imagebank("dewdropper")
  dpx, dpy = 20, 93
  maxhealth = 500
  def __init__(self, p):
    self.image = Dewdropper.imagebank.image()
    Structure.__init__(self, p)
    self.r /= 2.4
  def update(self):
    Structure.update(self)
    if not self.isdisabled():
      if random.randint(0, 800) == 0:
        d = Dewdrop((self.x, self.y))

class StarPlant (Structure):
  imagebank = imagebank.Imagebank("star-plant")
  dpx, dpy = 36, 32
  maxhealth = 800
  rstar = math.sqrt(0.1)
  group = pygame.sprite.Group()
  def __init__(self, p):
    self.image = StarPlant.imagebank.image()
    Structure.__init__(self, p)
    StarPlant.group.add(self)


class Dewdrop (Thing):
  image = None
  dpx, dpy = 10, 17
  group = pygame.sprite.Group()
  def __init__(self, p):
    if Dewdrop.image == None: self.setimage()
    self.rect = self.image.get_rect()
    self.r = self.rect.width / gamestate.kx / 2.2
    Thing.__init__(self, (p[0], p[1], 0.25))
    self.movable = True
    self.addshadow()
    for z in [0.2, 0.15, 0.1, 0.05, 0]:
      self.doq.append("self.move0((%f, %f, %f))" % (p[0], p[1], z))
    dx, dy = random.uniform(-0.3, 0.3), random.uniform(-0.3, 0.3)
    zs = parabola(0, 0.15, 0., 10) + parabola(0, 0.1, 0, 7) + [0,] * 4
    for j in range(len(zs)):
      x, y, z = p[0] + dx * j / len(zs), p[1] + dy * j / len(zs), zs[j]
      self.doq.append("self.move0((%f, %f, %f))" % (x, y, z))
    Dewdrop.group.add(self)
  def setimage(self):
    s = pygame.Surface((60, 60)).convert_alpha()
    s.fill((0,0,0,0))
    pygame.draw.ellipse(s, (128, 128, 255, 64), s.get_rect())
    pygame.draw.ellipse(s, (128, 128, 255, 255), s.get_rect(), 2)
    s = pygame.transform.smoothscale(s, (20, 20))
    Dewdrop.image = s
  def update(self):
    if len(self.doq):
      Thing.update(self)
      if not self.alive(): return
      if not self.moving():
        for c in Thing.tgroup.sprites():
          self.scootch(c)
    else:
      Thing.update(self)
  def apply(self):
    if not self.target: return
    if isinstance(self.target, Structure):
      self.target.addhealth(400)
      self.kill()

def parabola(y0, ymax, y1, n):
  f0, f1 = ymax - y0, ymax - y1
  if f0 == f1: x0 = 0.5
  else: x0 = (-f0 + math.sqrt(f0 * f1)) / (f1 - f0)
  return [ymax - f0*(x/x0/n - 1)**2 for x in range(n)]

class Acorn (Critter):
  imagebank = imagebank.Imagebank("acorn", 24)
  dpx, dpy = 12, 29
  frange = 0.5
  group = pygame.sprite.Group()
  def __init__(self, p):
    Critter.__init__(self, (0,0,0), Acorn.imagebank.image())
    self.group.add(self)
    self.gounavailable()
    if len(p) == 2: p = p + (0,)
    zs = parabola(0.35, 0.6, 0., 12) + parabola(0, 0.2, 0, 6) + [0, 0, 0]
    for j in range(len(zs)):
      x, y, z = p[0] * j / len(zs), p[1] * j / len(zs), zs[j]
      self.doq.append("self.move0((%f, %f, %f))" % (x, y, z))
    self.doq.append("self.move0((%f, %f, %f))" % p)
    self.doq.append("self.goavailable()")
    self.maxjump = 0.2
    self.speed = 0.03
  def nabnearestdrop(self):
    n, d2min = None, 100000
    for t in Thing.tgroup:
      if isinstance(t, Dewdrop) and t.leader == None:
        d2 = (self.x - t.x) ** 2 + (self.y - t.y) ** 2
        if d2 < d2min:
          n, d2min = t, d2
    if n:
      if d2min < (self.r + n.r) ** 2:
        self.nab(n)
      else:
        self.move((n.x, n.y))
        self.doq.append("self.nabnearestdrop()")
  def implant(self):
    for z in [0.05, 0.08, 0.09, 0.08, 0.05, 0]:
      self.doq.append("self.move0((%f, %f, %f))" % (self.x, self.y, z))
    for p in [10, 20, 30]:
      self.doq.append("self.image = Acorn.imagebank.image('under%d')" % p)
  def fertilize(self):
    self.gounavailable()
    self.implant()
    self.doq.append("self.kill()")
    for b in Structure.group:
      if (b.x - self.x) ** 2 + (b.y - self.y) ** 2 < self.frange ** 2:
        b.addhealth(100)
  def becomestarplant(self):
    if not canbuild((self.x, self.y), StarPlant): return False
    s = StarPlant((self.x, self.y))
    self.gounavailable()
    self.implant()
    self.doq.append("self.kill()")
  def becomedewdropper(self):
    if not canbuild((self.x, self.y), Dewdropper): return False
    s = Dewdropper((self.x, self.y))
    self.gounavailable()
    self.implant()
    self.doq.append("self.kill()")
  def becometurnip(self):
    self.gounavailable()
    t = Turnip((self.x, self.y))
    self.kill()
  def update(self):
    Critter.update(self)

class Turnip (Critter):
  imagebank = imagebank.Imagebank("turnip")
#  dpx, dpy = 12, 29
  def __init__(self, p):
    Critter.__init__(self, p, Turnip.imagebank.image())
    self.maxjump = 0.3
    self.speed = 0.015


# a semi-transparent oval that goes beneath things
class Shadow (pygame.sprite.Sprite):
  alpha = 40
  images = {}
  group = pygame.sprite.Group()
  def getimage(self, size):
    if size in Shadow.images: return Shadow.images[size]
    s = pygame.Surface((size,size/2)).convert_alpha()
    s.fill((0,0,0,0))
    pygame.draw.ellipse(s, (0,0,0,Shadow.alpha), s.get_rect())
    Shadow.images[size] = s
    return s
  def __init__(self, caster):
    pygame.sprite.Sprite.__init__(self)
    self.caster = caster
    self.size = caster.image.get_width()
    self.image = self.getimage(self.size)
    self.rect = self.image.get_rect()
    Shadow.group.add(self)
  def update(self):
    self.rect.center = gamestate.Pxy(self.caster.x, self.caster.y)

# A little arrow that tells you what units are selected
class Indicator(pygame.sprite.Sprite):
  image = None
  def __init__(self, target):
    pygame.sprite.Sprite.__init__(self)
    if not Indicator.image:
      Indicator.image = data.loadimg("arrow.png")
    self.rect = Indicator.image.get_rect()
    Thing.group.add(self)

# A progress indicator bar that goes above structures
class Ibar (pygame.sprite.Sprite):
  images = {}
  color0 = 255, 0, 0, 100
  color1 = 0, 255, 0, 100
  height = 7
  def getimage(self, w, wmax):
    if (w, wmax) in Ibar.images: return Ibar.images[(w, wmax)]
    s = pygame.Surface((wmax, self.height)).convert_alpha()
    s.fill(self.color1)
    if w < wmax-1:
      r = pygame.Rect(w, 1, wmax-2, self.height - 2)
      pygame.draw.rect(s, self.color0, r)
    Ibar.images[(w, wmax)] = s
    return s
  def __init__(self, wmax):
    pygame.sprite.Sprite.__init__(self)
    self.image = self.getimage(0, wmax)
    self.rect = self.image.get_rect()
    Thing.group.add(self)

class Twinkle (Thing):
  color = 128, 128, 255
  image = None
  def __init__(self, p):
    if Twinkle.image == None:
      Twinkle.image = pygame.Surface((2,2))
      Twinkle.image.fill(Twinkle.color)
    self.rect = Twinkle.image.get_rect()
    Thing.__init__(self, p)
    Thing.tgroup.remove(self)
    self.ps = []
    self.setpath()
  def setpath(self):
    if gamestate.coverage > 0.9:
      for j in range(40):
        f = math.exp(-j/8.)
        C, S = math.cos(j/10.), math.sin(j/10.)
        x, y = self.x * C + self.y * S, -self.x * C + self.y * S
        self.ps.append((x * f, y * f, j / 30.))
      return

    n, d2min = None, 100000
    for t in Thing.tgroup:
      if isinstance(t, StarPlant) and not t.isdisabled():
        d2 = (self.x - t.x) ** 2 + (self.y - t.y) ** 2
        if d2 < d2min: n, d2min = t, d2
    nearest, d = n, math.sqrt(d2min)
    if nearest and d < StarPlant.rstar:
      cx, cy = (nearest.x + self.x) / 2, (nearest.y + self.y) / 2
      rx, ry = (nearest.x - self.x) / 2, (nearest.y - self.y) / 2
      r = math.sqrt(rx ** 2 + ry ** 2)
      for p in range(12):
        theta = p * 3.1416 / 12
        c, s = math.cos(theta), math.sin(theta)
        x, y, z = cx - c * rx, cy - c * ry, s * r
        self.ps.append((x, y, z))
      return

    x, y = self.x, self.y
    for z in [j/20. for j in range(40)]:
      x += random.uniform(-0.02, 0.02)
      y += random.uniform(-0.02, 0.02)
      self.ps.append((x, y, z))
  def update(self):
    if len(self.ps): self.move0(self.ps.pop(0))
    Thing.update(self)
    if not len(self.ps): self.kill()

def canbuild(pos, c):
  x, y = pos
  if x**2 + y**2 > 1: return False
  for t in Structure.group:
    d2 = (x - t.x) ** 2 + (y - t.y) ** 2
    r = c.buildr + t.buildr
    if d2 < r ** 2: return False
  return True
      

def nearestthing(pos, c = Thing):
  if len(pos) == 3: x, y, z = pos
  else: x, y, z = pos + (0,)
  n, d2min = None, 100000
  for t in Thing.tgroup:
    if isinstance(t, c):
      d2 = (x - t.x) ** 2 + (y - t.y) ** 2 + (z - t.z) ** 2
      if d2 < d2min:
        n, d2min = t, d2
  return n, math.sqrt(d2min)


def update():
  Thing.group.update()
  Shadow.group.update()

def draw(surf):
  Shadow.group.draw(surf)
  Thing.group.draw(surf)
